/* Copyright (c) 2023 Skyward Experimental Rocketry
 * Author: Nicolò Caruso
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "MY_LIS3DSH.h"

#include <utils/Debug.h>

namespace Boardcore
{
/**
 * LIS3DSH class implementation
 */

/**
 * Constructors implementation
 */
MY_LIS3DSH::MY_LIS3DSH(SPIBusInterface& spiBus, miosix::GpioPin cs,
                       SPIBusConfig cfg, MY_LIS3DSHDefs::OutputDataRate odr,
                       MY_LIS3DSHDefs::BlockDataUpdate bdu,
                       MY_LIS3DSHDefs::FullScaleRange fs)
    : spi(spiBus, cs, cfg), odr(odr), bdu(bdu), fs(fs), initialized(false)
{
}

/*! Constructor without */

MY_LIS3DSH::MY_LIS3DSH(SPIBusInterface& spiBus, miosix::GpioPin cs,
                       MY_LIS3DSHDefs::OutputDataRate odr,
                       MY_LIS3DSHDefs::BlockDataUpdate bdu,
                       MY_LIS3DSHDefs::FullScaleRange fs)
    : MY_LIS3DSH(spiBus, cs, getDefaultSPIConfig(), odr, bdu, fs)
{
}

/**
 * Get default SPI configuration implementation
 * @return the SPIBusConfig for the given spi configuration
 */
SPIBusConfig MY_LIS3DSH::getDefaultSPIConfig()
{
    SPIBusConfig spiConfig;
    spiConfig.clockDivider =
        SPI::ClockDivider::DIV_256;     /*< Told by no idea by who...*/
    spiConfig.mode = SPI::Mode::MODE_3; /*< MODE 3 since MISO/MOSI on low clk*/
    spiConfig.byteOrder = SPI::Order::MSB_FIRST; /*< MSB... since SPI sends in
                                                    MSB from datasheet ...*/
    return spiConfig;
}
/**
 * Sensor initialization
 */
bool MY_LIS3DSH::init()
{
    /*! Already initialized */
    if (initialized)
    {
        lastError = SensorErrors::ALREADY_INIT;
        return false;
    }

    SPITransaction spiTr(spi);
    miosix::delayUs(100);

    uint8_t whoami =
        spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_WHO_AM_I);

    if (whoami != MY_LIS3DSHDefs::WHO_AM_I_ID)
    {
        LOG_ERR(logger,
                "Failed init. Cause: INVALID_WHO_AM_I Expecter value: {:X}. "
                "Actual value {:X}\n",
                MY_LIS3DSHDefs::WHO_AM_I_ID, whoami);
        TRACE("Bad WHOAMI... %d expected but... %d received!\n",
              MY_LIS3DSHDefs::WHO_AM_I_ID, whoami);
        lastError = SensorErrors::INVALID_WHOAMI;
        return false;
    }

    initialized = true;

    uint8_t ctrlReg4 = 0b0000'0000;
    uint8_t ctrlReg5 = 0b0000'0000;

    /*! Define the control register */
    ctrlReg4 = odr | bdu | MY_LIS3DSHDefs::CTRL_REG4_XEN |
               MY_LIS3DSHDefs::CTRL_REG4_YEN | MY_LIS3DSHDefs::CTRL_REG4_ZEN;

    spiTr.writeRegister(MY_LIS3DSHDefs::Registers::REG_CTRL_REG4, ctrlReg4);

    miosix::delayUs(10);
    /**
     * Verify that the ctrlReg4 is correctly initialized by comparing it with
     * the readed register*/
    initialized &= (ctrlReg4 == spiTr.readRegister(
                                    MY_LIS3DSHDefs::Registers::REG_CTRL_REG4));
    /*! Initializes CTRL_REG5 and see if it has been correctly initialized*/
    ctrlReg5 |= fs;
    spiTr.writeRegister(MY_LIS3DSHDefs::Registers::REG_CTRL_REG5, ctrlReg5);
    miosix::delayUs(10);
    initialized &= (ctrlReg5 == spiTr.readRegister(
                                    MY_LIS3DSHDefs::Registers::REG_CTRL_REG5));
    return initialized;
}

/**
 * Self test method.
 * It does perform sampling via self test to see if the full scale range is
 * correct.
 */
bool MY_LIS3DSH::selfTest()
{
    SPITransaction spiTr(spi);
    MY_LIS3DSHData dataTest, dataNormal, accumulatorTest, accumulatorNormal;
    bool test      = true;
    int i          = 0;
    uint8_t ctrl_5 = 0b0000'0000, test_ctrl_5 = 0b0000'0000;
    ctrl_5      = spiTr.readRegister(MY_LIS3DSHDefs::REG_CTRL_REG5);
    test_ctrl_5 = ctrl_5 | MY_LIS3DSHDefs::ST_POSITIVE;
    miosix::delayMs(1000);

    /*! Performs a clean sampling*/
    for (i = 0; i < 20; i++)
    {
        miosix::delayMs(10);
        dataNormal = sampleImpl();
        accumulatorNormal.accelerationX += dataNormal.accelerationX;
        accumulatorNormal.accelerationY += dataNormal.accelerationY;
        accumulatorNormal.accelerationZ += dataNormal.accelerationZ;
    }

    /*! Sets the control register 5 to do a self test*/
    spiTr.writeRegister(MY_LIS3DSHDefs::REG_CTRL_REG5, test_ctrl_5);
    miosix::delayUs(1000);

    /*! Performs a self test sampling*/
    for (i = 0; i < 20; i++)
    {
        miosix::delayMs(10);
        dataTest = sampleImpl();
        accumulatorTest.accelerationX += dataTest.accelerationX;
        accumulatorTest.accelerationY += dataTest.accelerationY;
        accumulatorTest.accelerationZ += dataTest.accelerationZ;
    }

    /*! Resets the control register 5*/
    spiTr.writeRegister(MY_LIS3DSHDefs::REG_CTRL_REG5, ctrl_5);

    /*! Self test for X, Y is of 140[mg]*/
    test &= ((accumulatorTest.accelerationX / i) -
                 (accumulatorNormal.accelerationX / i) >
             0.140 - 2);
    test &= ((accumulatorTest.accelerationX / i) -
                 (accumulatorNormal.accelerationX / i) <
             0.140 + 2);
    test &= ((accumulatorTest.accelerationY / i) -
                 (accumulatorNormal.accelerationY / i) >
             0.140 - 2);
    test &= ((accumulatorTest.accelerationY / i) -
                 (accumulatorNormal.accelerationY / i) <
             0.140 + 2);

    /*! Self test for Z is of 590[mg]*/
    test &= ((accumulatorTest.accelerationZ / i) -
                 (accumulatorNormal.accelerationZ / i) >
             0.590 - 2);
    test &= ((accumulatorTest.accelerationZ / i) -
                 (accumulatorNormal.accelerationZ / i) <
             0.590 + 2);
    return test;
}

MY_LIS3DSHData MY_LIS3DSH::sampleImpl()
{

    /*! Last sample timestamp and variables for the values */
    uint64_t lastSampleTimestamp = TimestampTimer::getTimestamp();

    float x = 0;
    float y = 0;
    float z = 0;

    /*! Read the data registers for (X, Y, Z)*/
    {
        /**
         * SPI transaction and buffer of 6 integers (2x3 for the int_16
         * registers)*/
        SPITransaction spiTr(spi);

        /*! Read from STATUS REGISTER */
        uint8_t status =
            spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_STATUS_REG);
        /*! Concatenates the 8-bit read, into the buffer for 16-bit int values*/
        /*! Order given by the STATUS_REG low and high output registers*/
        int16_t regX =
            spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_OUT_X_H) << 8 |
            spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_OUT_X_L);
        int16_t regY =
            spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_OUT_Y_H) << 8 |
            spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_OUT_Y_L);
        int16_t regZ =
            spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_OUT_Z_H) << 8 |
            spiTr.readRegister(MY_LIS3DSHDefs::Registers::REG_OUT_Z_L);

        /*! In case of no new data available, then just return last one*/
        if (!(status & MY_LIS3DSHDefs::STATUS_REG_XYZDA))
        {
            lastError = SensorErrors::NO_NEW_DATA;
            TRACE("No new data available.");
            return lastSample;
        }

        /*! Sensitivity based on the FS */
        float sensitivity;
        sensitivity = MY_LIS3DSHDefs::SENSITIVITY_VALUES[fs >> 3];

        /*! Samples to [g]. Read register * sensitivity in [mg] /1000 to have
         * [g]*/
        x = regX * sensitivity / 1000;
        y = regY * sensitivity / 1000;
        z = regZ * sensitivity / 1000;
    }
    return MY_LIS3DSHData(lastSampleTimestamp, x, y, z);
}
}  // namespace Boardcore